#!/bin/bash
########################################################################
# build_tensorflow.sh
#
# Build tensorflow and package for UPS.
#
# Adapted from build.sh.template provided by ssibuildshims v2_09_00.
########################################################################

########################################################################
# Non-customize-able boilerplate: skip to "Customize-able functions."
########################################################################

####################################
# Useful variables.
prog=${BASH_SOURCE[0]##*/}
thisdir=$(cd "${BASH_SOURCE%/*}" && pwd -P)

##################
# Local (dumb) ssi_die() function for use until we source ssi_functions.
ssi_die() {
  printf "FATAL ERROR: ${*}\n" 1>&2; exit 1
}
##################

##################
# Basic usage function for use by ensure_buildshims until we have enough
# information to provide something more specific. Do NOT customize!
build_script_usage() {
    cat <<EOF
USAGE: ${prog} <ssi-msg-config-options> [--] <product-dir> <qualifier-args> <maketar>
       ${prog} -h|--help [<product-dir>]

  Build ${package} and package for UPS.

OPTIONS

  -h|--help

    Help. Supply <product-dir> for specific, detailed information on
    invoking ${prog}.

  <ssi-msg-config-options>

    Supply <product-dir> for more information.

ARGUMENTS

  <product-dir>

    The top-level UPS directory in which the source code distribution
    was unpacked. When specified with -h|--help, this enables us to
    provide more specific help about invoking ${prog}.

  <qualifier-args>
  <maketar>

    Supply <product-dir> for more information.

EOF
}
##################

##################
# Make sure we have UPS and ssibuildshims set up.
ensure_ssibuildshims() {
  if ! (( $# )) || [ "${@: -1}" = "-h" ] || [ "${@: -1}" = "--help" ]; then
    build_script_usage
    exit 1
  fi
  # Find <product-dir> from our arguments:
  local arg
  for arg in "${@}"; do
    [[ "$arg" ==  -* ]] && continue;
    product_dir="${arg}"; break
  done
  if [ -z "${product_dir}" ]; then
    # Bad usage.
    ssi_die "Bad <product-dir> ${product_dir}\n$(build_script_usage)"
    exit 2
  elif ! { [ -d "${product_dir}/.upsfiles" ] && [ -w "${product_dir}" ]; }; then
    # Something wrong with <product-dir>.
    ssi_die "<product-dir> not a valid, writable unified UPS directory: " \
      "${product_dir}\n$(build_script_usage)"
  fi
  # Promising...
  local abs_product_dir="$(cd "${product_dir}" 2>/dev/null 2>&1 && pwd -P)"
  (( $? )) && ssi_die "unable to cd to <product-dir>: ${product_dir}\n" \
    "$(build_script_usage)"
  # Find our own bootstraps and yank 'em.
  local ssibuildshims_topdir="${product_dir}/ssibuildshims/${ssibuildshims_version}"
  [ -f "${ssibuildshims_topdir}/bin/ssi_functions" ] && \
    source "${ssibuildshims_topdir}/bin/ssi_functions" >/dev/null 2>&1 && \
    ensure_ups "${product_dir}" && \
    setup ssibuildshims ${ssibuildshims_version} -z "${product_dir}" || \
    ssi_die "unable to find ssibuildshims ${ssibuildshims_version} in
${product_dir} and set it up with UPS."
  # Absolute is safer.
  absolute_path -- "${product_dir}" product_dir
}
##################

########################################################################
# Customize-able functions.
########################################################################

##################
# Ascertain whether we built and installed correctly. Add more /
# different calls to verify_in_filesystem as appropriate. Arguments
# (e.g. -q) are passed through to verify_in_filesystem.
is_built() { # *C*
  ${trace} verify_in_filesystem -d "${@}" "${pkgdir}/lib"{64,}
}
##################

##################
# If a distribution archive is wanted and the installation is successful
# according to is_built(), then make it. Any arguments (e.g. -q) are
# passed through both to is_built() and make_basic_distribution().
#
# If make_basic_distribution() finds that the product is declared
# correctly and attempts to make the distribution, then it will cause us
# to exit with the status code of that operation. If not, it returns an
# error code.
#
# See make_basic_distribution -h for useful options to add to
# mbd_opts(). *C*
maybe_make_distribution() {
  if [ "${maketar}" = "tar" ]; then
    # Customize mbd_opts. *C*
    local mbd_opts=()
    ${trace} is_built "$@" && \
      ${trace} make_basic_distribution "${@}" "${mbd_opts[@]}"
  fi
}
##################

########################################################################
# Package-specific functions.
########################################################################

# ...

########################################################################
# Main.
########################################################################

####################################
# Useful variables. *C*
#trace=maybe_xtrace # Could set to xtrace instead for unconditional.
(( ssi_trace = 1 )) # Trace certain commands by default.
####################################

####################################
# Package name and version information. *C*
. ${thisdir}/version_info.sh || \
  ssi_die "unable to load version info from ${thisdir}/version_info"
####################################

####################################
# Handle options and arguments.

# Set up ssibuildshims and UPS / product_dir as early as possible.
ensure_ssibuildshims "${@}"

##################
# Now parse options and arguments, setting basequal, extraqual, maketar,
# etc.
#
# Use verify_quals() to parse command line options, verify that the user
# has provided qualifiers supported by this build script, and set
# expected variables such as basequal, extraqual, cqual, etc.
#
# Customize the extra usage text as appropriate. Note that DEBUG_PATCHES
# and VERBOSE_PATCHES are honored by prepare_source() and
# patch_source(). DEBUG_(CONFIG|BUILD|TEST|INSTALL) should be accounted
# for (or not) by the relevant command in this script.
#
# See verify_quals -h for options and examples.
verify_quals \
  --cqual={e19,e20,c7} \
  --pqual={p383b,p392} \
  --parse --usage-file=- -- "$@" <<EOF || exit # *C*
ENVIRONMENT

  DEBUG_(CONFIG|BUILD|TEST|INSTALL)

    Verbose output from the corresponding step.

  DEBUG_PATCHES
  VERBOSE_PATCHES

    Command tracing and increased verbosity for source unpacking and patching
    operations. DEBUG_PATCHES will cause exit after patches have been applied.
EOF
##################

##################
# Define globals: pkgdir, sourcedir, blddir, fullqual, etc. See
# define_basics -h for steering options.
${trace} define_basics IN_SOURCE NO_TYPE TRACK_PROGRESS || exit # *C*
##################

####################################
# Declare ourselves privately, and then set ourselves up to allow setup
# of dependencies. *C*
#
# Comment or delete if not required:
${trace} fake_declare_product "${ssi_args[@]}" -x -- || exit
####################################

####################################
# Build-only dependencies. *C*
#
# e.g.
#   ${trace} setup_cmake --latest={--ge,3.15} || exit
#      (version requirement necessary for cmake --install).
# AND/OR
#   ${trace} setup_ninja --optional && (( want_ninja = 1 ))

# Define versions of build-time-only products to be set up.
    pybind11_default_version=v2_5_0a
    eigen_default_version=v3_3_9a
    protobuf_default_version=v3_12_3a
    bazel_default_version=v3_1_0
    swig_default_version=v4_0_2

# Build tools.
${trace} setup_prod bazel -- -q+  || exit 
${trace} setup_prod swig -- -q+${cqual}:prof  || exit 
${trace} setup_prod protobuf -- -q+${cqual}  || exit 
##################
# These packages are either header-only or expose neither dependency nor
# defined symbols. Default versions are determined above.
      ${trace} setup_prod eigen -- -q + && \
      ${trace} setup_prod pybind11 -- -q + || exit

##################
# Define globals: pkgdir, sourcedir, blddir, fullqual, etc. See
####################################

# Print active products.
echo
ups active
echo

# Possible short-circuit if we're already built and declared.
${trace} maybe_make_distribution --silent

ssi_info "building ${package} for ${OS}-${plat}-${qualdir} (flavor ${pkgflvr})"

# Ensure directories: pkgdir, sourcedir, blddir.
${trace} init_pkg_dirs || exit

####################################
# Prepare source for building. *C*

${trace} cd "${sourcedir}" \
  || ssi_die "unable to cd to source directory ${sourcedir}"

case ${cqual} in 
  e21)
    patches=(v2.3.1-c++20.patch bfloat16.patch)
    ;;
  *)
    patches=(v2.3.1.patch bfloat16.patch)
    ;;
esac

${trace} cd "${sourcedir}" \
  || ssi_die "unable to cd to source directory ${sourcedir}"

# Source tarballs and patches. See prepare_source -h / patch_source -h
# for options.
${trace} prepare_source "${ssi_args[@]}" -P "${patchdir}" \
  "${tardir}/${pkgtarfile}" "${patches[@]}" || exit

# Bail early if DEBUG_PATCHES is set.
[ -n "${DEBUG_PATCHES}" ] && exit 1
####################################

####################################
# Prepare to build. *C*
#
# Set configure / compile flags, environment variables, etc.

declare -a cflg cxxflg

extra_cmake_args=(--)

case ${cqual} in
  e19) cc=gcc; cxx=g++; cstd=11; cxxstd=17
    cxxflg+=(-Wno-deprecated-copy -frecord-gcc-switches)
    ;;   
  e20) cc=gcc; cxx=g++; cstd=11; cxxstd=17
    cxxflg+=(-Wno-deprecated-copy -frecord-gcc-switches)
    ;;
  e21) cc=gcc; cxx=g++; cstd=11; cxxstd=20
    cxxflg+=(-fno-char8_t -Wno-deprecated-copy -frecord-gcc-switches)
    ;;
  c7) cc=clang; cxx=clang++; cstd=11; cxxstd=17
    cxxflg+=(-D_LIBCPP_ENABLE_CXX17_REMOVED_FEATURES=1 -grecord-gcc-switches)
    ;;
  *) ssi_die -I "accepted unknown cqual=${cqual}"
esac

common_flags=( -fno-math-errno
  -Wno-deprecated-declarations
  -Wno-parentheses
  -msse3
)

cflg+=("${common_flags[@]}"
  ${cstd:+-std=c${cstd}}
)

cxxflg+=("${common_flags[@]}"
  -Wno-redundant-move
  ${cxxstd:+-std=c++${cxxstd}}
)

##################
# Configure parallelism. *C*
ncores=$(ncores)

# For cmake --build:
#
# [[ -n "${CMAKE_BUILD_PARALLEL_LEVEL}" ]] || \
#   export CMAKE_BUILD_PARALLEL_LEVEL=${ncores}
#
# For ctest:
#
# [[ -n "${CTEST_PARALLEL_LEVEL}" ]] || \
#   export CTEST_PARALLEL_LEVEL=${ncores}
##################

compile_options=()

##################
# Conditional compilation options.
(( DEBUG_CONFIG )) || (( DEBUG_BUILD )) && compile_options+=(VERBOSE=1)
####################################

####################################
# Configure, build, and install.
#
# Use ${blddir}/${pkgtarname} for in-source builds, ${blddir} otherwise.

#${trace} cd "${blddir}" || \
${trace} cd "${blddir}/${pkgtarname}" || \
  ssi_die "unable to cd to build directory ${blddir}/${pkgtarname}"

##################
# Configure. *C*

# e.g.
#   ${trace} ${sourcedir}/${pkgtarname}/configure \
#     --prefix="${pkgdir}" ... || ssi_die "configure failed"
# OR
#   ${trace} cmake <opts> ${DEBUG_CONFIG:+--verbose} \
#     "${sourcedir}/${pkgtarname}" || \
#     ssi_die "configure failed"

export PYTHON_BIN_PATH="$(which python)"
export USE_DEFAULT_PYTHON_LIB_PATH=1
case ${cqual} in
  e*) 
    export GCC_HOST_COMPILER_PATH="$(which gcc)"
    export CC="$(which gcc)"
    export CXX="$(which g++)"
    ;;
  c*)
    export GCC_HOST_COMPILER_PATH="$(which clang)"
    export CC="$(which clang)"
    export CXX="$(which clang++)"
    ;;
  *)  ssi_die -I "accepted unknown cqual=${cqual}"
esac
cc_opt_flags=("${cflg[@]/#/--copt=}")

echo cc_opt_flags=${cc_opt_flags[@]}

cxx_opt_flags=("${cxxflg[@]/#/--cxxopt=}")
echo cxx_opt_flags=${cxx_opt_flags[@]}

export CC_OPT_FLAGS="-msse3"
export CXX_OPT_FLAGS="${cxxstd:+-std=c++${cxxstd}}"
export TF_NEED_JEMALLOC=0
export TF_NEED_HDFS=0
export TF_NEED_GCP=0
export TF_ENABLE_XLA=0
export TF_NEED_OPENCL=0
export TF_NEED_CUDA=0
export TF_NEED_VERBS=0
export TF_NEED_MKL=0
export TF_NEED_MPI=0
export TF_NEED_S3=0
export TF_NEED_GDR=0
export TF_NEED_OPENCL_SYCL=0
export TF_SET_ANDROID_WORKSPACE=false
export TF_NEED_KAFKA=false
export TF_NEED_AWS=0
export TF_DOWNLOAD_CLANG=0
export TF_NEED_IGNITE=0
export TF_NEED_ROCM=0
export TF_NEED_TENSORRT=0
export TEST_TMPDIR="$PWD/build"
export TF_CMS_EXTERNALS="$PWD/cms_externals.txt"

echo "com_google_protobuf:${PROTOBUF_FQ_DIR}" >> ${TF_CMS_EXTERNALS}
echo "org_sqlite:${SQLITE_FQ_DIR}"                  >> ${TF_CMS_EXTERNALS}
echo "swig:${SWIG_FQ_DIR}"                          >> ${TF_CMS_EXTERNALS}

export TF_SYSTEM_LIBS=$(cat ${TF_CMS_EXTERNALS} | sed 's|:.*||' | tr "\n" "," | sed 's|,*$||')

## install some python packages needed by pip package build
mkdir -p $PWD/lib/$PYTHON_LIBDIR/site_packages
pip install -t $PWD/lib/$PYTHON_LIBDIR/site_packages numpy protobuf keras-applications keras-preprocessing
export PYTHONPATH=$PWD/lib/$PYTHON_LIBDIR/site_packages:$PYTHONPATH


./configure

##################

##################
# Build. *C*

echo "build --action_env LD_LIBRARY_PATH=\"${LD_LIBRARY_PATH}\"" >>.tf_configure.bazelrc
# define bazel options
BAZEL_OPTS="--batch --output_user_root ../build build -s --verbose_failures -j HOST_CPUS*0.5"
# adding prof options
BAZEL_OPTS="$BAZEL_OPTS --config=opt ${cc_opt_flags[@]} ${cxx_opt_flags[@]}"
# other options
BAZEL_OPTS="$BAZEL_OPTS --config=noaws --config=nogcp --config=nohdfs --config=nonccl"
BAZEL_OPTS="$BAZEL_OPTS --action_env=SWIG_LIB=${SWIG_LIB}"
BAZEL_OPTS="$BAZEL_OPTS --action_env=LD_LIBRARY_PATH=${LD_LIBRARY_PATH}"
BAZEL_OPTS="$BAZEL_OPTS --distinct_host_configuration=false"

case ${ctype} in 
  c*) BAZEL_OPTS="$BAZEL_OPTS  --config=libc++";;
  *) ;;
esac

# build tensorflow library targets
bazel $BAZEL_OPTS //tensorflow:tensorflow  || exit
bazel $BAZEL_OPTS //tensorflow:tensorflow_cc || exit
bazel $BAZEL_OPTS //tensorflow/tools/graph_transforms:transform_graph || exit
bazel $BAZEL_OPTS //tensorflow/compiler/tf2xla:tf2xla || exit
bazel $BAZEL_OPTS //tensorflow/compiler/xla:cpu_function_runtime || exit
bazel $BAZEL_OPTS //tensorflow/compiler/xla:executable_run_options || exit
bazel $BAZEL_OPTS //tensorflow/compiler/tf2xla:xla_compiled_cpu_function || exit
bazel $BAZEL_OPTS //tensorflow/core/profiler || exit
bazel $BAZEL_OPTS //tensorflow:install_headers || exit
# build tensorflow python targets
bazel $BAZEL_OPTS //tensorflow/tools/pip_package:build_pip_package || exit

# rebuild *.pb.{h|cc} files using the external protobuf compiler
chmod -R a+rwX $PWD/bazel-bin/tensorflow/include
for f in $(find tensorflow -name "*.proto")
do
    protoc --cpp_out=$PWD/bazel-bin/tensorflow/include $f
done

# define and create empty target directories
bindir="$pkgdir/bin"
incdir="$pkgdir/include"
libdir="$pkgdir/lib"
rm -rf $bindir $incdir $libdir
mkdir -p $bindir $incdir $libdir

# copy targets
srcdir="${PWD}/bazel-bin/tensorflow"

cp -pdv $srcdir/libtensorflow*.so* $libdir/
cp -pdv $srcdir/compiler/tf2xla/lib*.so* $libdir/
cp -pdv $srcdir/compiler/xla/lib*.so* $libdir/

for l in tensorflow_cc tensorflow_framework tensorflow ; do
  # check if the actual lib exists
  [ -f $libdir/lib${l}.so.2.3.1 ] || ssi_die "missing expected library $libdir/lib${l}.so.2.3.1"

  # link from majorversion to realversion
  rm -f $libdir/lib${l}.so.2
  ln -s lib${l}.so.2.3.1 $libdir/lib${l}.so.2

  # link from default lib to majorversion
  rm -f $libdir/lib${l}.so
  ln -s lib${l}.so.2 $libdir/lib${l}.so
done

#cp -p $srcdir/compiler/aot/tfcompile $bindir
for name in tensorflow absl re2 third_party ; do
    cp -r -p $srcdir/include/$name $incdir
done

# copy headers from downloaded dependencies
copy_headers() {
  for header_file in $(find $1/$2 -name *.h | sed "s|$1/||") ; do
    header_dir="${incdir}/$(dirname ${header_file})"
    mkdir -p "${header_dir}"
    cp -pdv "${header_file}" "${header_dir}/"
  done
}
copy_headers "${PWD}" tensorflow/compiler
copy_headers "${PWD}" tensorflow/core/profiler/internal
copy_headers "${PWD}" tensorflow/core/profiler/lib

# build pip whl file for distributing tensorflow python package 
bazel-bin/tensorflow/tools/pip_package/build_pip_package ${pkgdir}

# e.g.
#   ${trace} cmake --build "${blddir}" ${DEBUG_BUILD:+--verbose} || \
#     ssi_die "build failed"
##################

##################
# Test. *C*

# e.g.
# (( WANT_TESTS )) && \
#   { ${trace} ctest --output-on-failure || ssi_die "tests failed"; }
##################

##################
# Install. *C*

# e.g.
#   ${trace} cmake --install "${blddir}" || ssi_die "install failed"
##################

ssi_info "finished building ${package} ${pkgver}"

####################################
# Pre-declaration processing. *C*
#
# Move things, install from contrib, clean build products from in-source
# builds, etc.
####################################

##################
# Declare and set up the installed package. *C*
#
# Add -c for current declaration if required.
${trace} declare_product -x -- || exit
##################

####################################
# Post-declaration processing. *C*
#
# Operations requiring the *final* declaration and setup (as opposed to
# the pre-build fake declaration and setup). This should be rare.
####################################

#################
# Verify installation and make distribution archive. *C*
#
# To customize verification, see is_built(), since it is also used by
# maybe_make_distribution(). See the documentation for those functions
# (above) for customization opportunities.

# Add a message to ssi_build_failed if appropriate.
${trace} is_built && ${trace} ssi_build_completed && \
  ssi_info "${package} is installed at ${pkgdir}" && \
  ${trace} maybe_make_distribution || \
  ssi_build_failed # *C*
########################################################################
# Nothing beyond this point.
########################################################################
